from .quantity import Quantity
from .string_conv import from_string, from_string_with_units, to_string
from .maths import add, greater, greater_equal, isclose, less
from .maths import less_equal, log10, neg, pow, sub
from .maths import absolute as abs

__all__ = ['Quantity',
           'from_string', 'to_string', 'from_string_with_units',
           'abs', 'add', 'greater', 'greater_equal', 'isclose', 'less',
           'less_equal', 'log10', 'neg', 'pow', 'sub']



from __future__ import annotations
from typing import Generator, List, Optional, TextIO, Tuple, Union
import gzip
from enum import Flag
from pathlib import Path
import Bio.PDB as biopdb
from Bio.PDB.Structure import Structure as bpStructure
from Bio.PDB.Chain import Chain as bpChain
from Bio.PDB.Residue import Residue as bpResidue
from Bio.PDB.Polypeptide import standard_aa_names
from Bio.PDB.Polypeptide import three_to_one
import oddt


class ResidueType(Flag):
    RESIDUE = 1
    HETERORESIDUE = 2
    WATER = 4
    ALL = RESIDUE | HETERORESIDUE | WATER

    def get_id_str(self):
        return ' ' if ResidueType.RESIDUE in self else '' + \
            'H' if ResidueType.HETERORESIDUE in self else '' + \
            'W' if ResidueType.HETERORESIDUE in self else ''


def _save(entity: Union[Protein, Chain, Residue],
          filename: Union[str, Path, TextIO]) -> None:
    if isinstance(filename, Path):
        filename = str(filename)

    io = biopdb.PDBIO()
    io.set_structure(entity.as_biopython())
    io.save(filename)


class Protein():
    def __init__(self, id: str):
        self.id = id
        self.filename: Optional[Path] = None
        self._biopython: Optional[bpStructure] = None
        self._oddt: Optional[oddt.toolkit.Molecule] = None

    @property
    def name(self) -> str:
        return self.id

    def as_biopython(self) -> bpStructure:
        if not self._biopython:
            if self.filename:
                file_open = gzip.open if self.filename.suffix == '.gz' \
                    else open

                with file_open(self.filename, 'rt', encoding='ascii') as file:
                    parser = biopdb.PDBParser()
                    self._biopython = parser.get_structure(None, file)
            else:
                raise Exception('No filename is specified.')

        return self._biopython

    def as_oddt(self) -> oddt.toolkit.Molecule:
        if not self._oddt:
            if self.filename:
                mol = next(oddt.toolkit.readfile('pdb', str(self.filename)))
                mol.protein = True

                self._oddt = mol
            else:
                raise Exception('No filename is specified.')

        return self._oddt

    def get_chain(self, chain_id: str) -> Chain:
        return Chain(self, chain_id)

    def get_chains(self) -> Generator[Chain, None, None]:
        chains = self.as_biopython().get_chains()
        return (Chain(self, c.id, biopython=c) for c in chains)

    def save(self, filename: Union[str, Path, TextIO]):
        _save(self, filename)

    @staticmethod
    def from_file(filename: Union[str, Path]) -> Protein:
        if isinstance(filename, str):
            filename = Path(filename)

        filename_unzip = filename.stem if filename.suffix in ['.gz'] \
            else filename
        name = Path(filename_unzip).stem

        protein = Protein(name)
        protein.filename = filename

        return protein


class Chain():
    def __init__(self, protein: Protein, id: str,
                 biopython: Optional[bpChain] = None):
        self.protein = protein
        self.id = id
        self._biopython = biopython

    @property
    def name(self) -> str:
        return f'{self.protein.name}_{self.id}'

    def as_biopython(self) -> bpChain:
        if not self._biopython:
            structure = self.protein.as_biopython()
            chains = [c for c in structure.get_chains() if c.id == self.id]
            self._biopython = chains[0]

        return self._biopython

    def get_residue(self,
                    residue_number: int) -> Optional[Residue]:
        for r in self.as_biopython().get_residues():
            if r.id[1] == residue_number:
                return Residue(self, r.id, biopython=r)
        return None

    def get_residues(self,
                     residue_type: ResidueType = ResidueType.ALL,
                     residue_names: Optional[List[str]] = None,
                     ) -> Generator[Residue, None, None]:
        def filter(r: bpResidue):
            return r.id[0][0] in residue_ids \
                and (r.resname in residue_names if residue_names else True)

        residue_ids = residue_type.get_id_str()
        residues = self.as_biopython().get_residues()
        return (Residue(self, r.id, biopython=r) for r in residues
                if filter(r))

    def get_sequence(self):
        def _three_to_one(s: str):
            return three_to_one(s) if s in standard_aa_names else '?'

        residues = self.as_biopython().get_residues()
        sequence = [_three_to_one(residue.get_resname())
                    for residue in residues]
        sequence = ''.join(sequence)
        return sequence

    def save(self, filename: Union[str, Path, TextIO]):
        _save(self, filename)

    def __repr__(self):
        return f'<Chain id={self.id}>'


class Residue():
    def __init__(self, chain: Chain, id: Tuple[str, int, str],
                 biopython: Optional[bpResidue] = None):
        self.chain = chain
        self.id = id
        self._biopython = biopython

    @property
    def name(self) -> str:
        return f'{self.chain.name}_{self.residue_id}'

    @property
    def num_atoms(self):
        return sum(1 for _ in self.as_biopython().get_atoms())

    @property
    def residue_id(self):
        return f'{self.residue_name}{self.id[1]}'

    @property
    def residue_name(self):
        return self.as_biopython().resname

    def as_biopython(self) -> bpResidue:
        if not self._biopython:
            chain = self.chain.as_biopython()
            residues = [r for r in chain.get_residues() if r.id == self.id]
            self._biopython = residues[0]

        return self._biopython

    def save(self, filename: Union[str, Path, TextIO]):
        _save(self, filename)

    def __repr__(self):
        return f'<Residue id={self.id}>'



from tensorflow.keras import Input
from chemmltoolkit.tensorflow.graph import TensorGraphShape
from chemmltoolkit.tensorflow.graph.tensorGraph import NODE_FEATURES
from chemmltoolkit.tensorflow.graph.tensorGraph import EDGE_FEATURES


def TensorGraphInput(shape: TensorGraphShape,
                     batch_size=None,
                     name: str = 'graph'):
    """Returns a Keras input for TensorGraphs.

    This function simply returns a dictionary with keras Input elements
    for each tensor in the input TensorGraph.

    Arguments:
        shape: The shape of the input graph.
        batch_size: Optional static batch size (integer).
        name: An optional prefix for the names of the Input layers. The
            default value is 'graph'.
    """
    return {
        NODE_FEATURES: Input(
            shape=(shape.num_nodes, shape.node_dims),
            batch_size=batch_size,
            name=f'{name}_node_features'),
        EDGE_FEATURES: Input(
            shape=(shape.edge_dims, shape.num_nodes, shape.num_nodes),
            batch_size=batch_size,
            name=f'{name}_edge_features')
    }



from chemmltoolkit.tensorflow.layers.graphConv import GraphConv
from chemmltoolkit.tensorflow.layers.oneHotEncoding import OneHotEncoding
from chemmltoolkit.tensorflow.layers.shiftTensor import ShiftTensor

__all__ = ['GraphConv', 'OneHotEncoding', 'ShiftTensor']



from typing import Generator, List, Optional, Union
import math
import numpy as np
from numpy.typing import NDArray
from . import Protein, Residue
import oddt
from oddt.interactions import hbonds


def get_covalent_residue(residue: Residue,
                         convalent_residues: List[str] = ['CYS'],
                         distance_cutoff: float = 2.0
                         ) -> Optional[Residue]:
    distance_squared = distance_cutoff ** 2

    ligand_coords = [atom.coord for atom in residue.as_biopython().get_atoms()]

    for residue_cov in residue.chain.get_residues(
            residue_names=convalent_residues):
        residue_cov_biopython = residue_cov.as_biopython()
        if residue_cov_biopython == residue.as_biopython():
            continue

        for atom in residue_cov_biopython.get_atoms():
            atom_coord = atom.coord
            dists_squared = [np.sum((c - atom_coord)**2)
                             for c in ligand_coords]
            for dist in dists_squared:
                if dist < distance_squared:
                    return residue_cov

    return None


def contact_distance(a: List[Union[Residue, NDArray[np.float32]]],
                     b: List[Union[Residue, NDArray[np.float32]]]
                     ) -> float:
    def _distance_squared(coords_a: NDArray[np.float32],
                          coords_b: NDArray[np.float32]) -> float:
        return np.sum((coords_a-coords_b)**2)

    from_coords = list(_get_atom_coordinates(a))
    return math.sqrt(min(
        min(_distance_squared(c, to_coord) for c in from_coords)
        for to_coord in _get_atom_coordinates(b)))


def get_interactions(protein: Protein,
                     ligand: oddt.toolkit.Molecule,
                     strict_only: bool = True):
    protein_oddt = protein.as_oddt()
    prot_atoms, lig_atoms, strict = hbonds(protein_oddt, ligand)
    chain = next(protein.get_chains())

    if strict_only:
        prot_atoms = prot_atoms[strict]
        lig_atoms = lig_atoms[strict]

    interactions = []

    for prot_atom, lig_atom in zip(prot_atoms, lig_atoms):
        # NB: Filter out interactions not to protein residues
        #     e.g. ligand-self, ligand-solvent
        if prot_atom['resnum'] in protein_oddt.res_dict['resnum']:
            residue = chain.get_residue(prot_atom['resnum'])

            if lig_atom['isacceptor']:
                interaction_type = 'A'
            elif lig_atom['isdonor']:
                interaction_type = 'D'
            else:
                interaction_type = 'X'

            interaction = (residue, prot_atom['isbackbone'], interaction_type)
            interactions.append(interaction)

    return interactions


def _get_atom_coordinates(items: List[Union[Residue, NDArray[np.float32]]]
                          ) -> Generator[NDArray[np.float32], None, None]:
    for item in items:
        if isinstance(item, Residue):
            for atom in item.as_biopython().get_atoms():
                yield atom.coord
        else:
            yield item



from __future__ import annotations
from typing import Dict
from chemmltoolkit.features.atomFeaturiser import AtomFeaturiser
from chemmltoolkit.features.bondFeaturiser import BondFeaturiser
from rdkit.Chem import Mol
import torch
import dgl


class GraphFeaturiser():
    """Generator for graph-based features.
    Args:
        features: A list of features to generate.
    """

    def __init__(self: GraphFeaturiser,
                 atom_featurisers: Dict[str, AtomFeaturiser],
                 bond_featurisers: Dict[str, BondFeaturiser],
                 add_self_loops: bool = True):
        self.atom_featurisers = atom_featurisers
        self.bond_featurisers = bond_featurisers
        self.add_self_loops = add_self_loops

    def to_dgl(self: GraphFeaturiser, mol: Mol) -> dgl.DGLGraph:
        """Generates a DGL graph from a molecule.

        Args:
            mol: The molecule to featurise.

        Returns:
            A DGL graph of the featurised molecule.
        """
        num_atoms = mol.GetNumAtoms()
        bonds = mol.GetBonds()
        bond_from = [bond.GetBeginAtomIdx() for bond in bonds]
        bond_to = [bond.GetEndAtomIdx() for bond in bonds]

        g = dgl.graph((torch.tensor(bond_from), torch.tensor(
            bond_to)), num_nodes=num_atoms)

        for key, atom_featuriser in self.atom_featurisers.items():
            atom_features = atom_featuriser.process_molecule(mol)
            g.ndata[key] = torch.tensor(atom_features, dtype=torch.float)

        for key, bond_featuriser in self.bond_featurisers.items():
            bond_features = [bond_featuriser.process_bond(
                bond) for bond in bonds]
            g.edata[key] = torch.tensor(bond_features, dtype=torch.float)

        g = dgl.add_reverse_edges(g, copy_edata=True)

        if self.add_self_loops:
            g = dgl.add_self_loop(g)

        return g



from pathlib import Path
from typing import Dict, Optional, Union
from . import Residue
from chemmltoolkit.utils.data_utils import get_file
import csv
from io import StringIO
import openbabel.pybel as pybel
from rdkit import Chem
from rdkit.Chem.AllChem import AssignBondOrdersFromTemplate
from rdkit.Chem import Mol
import oddt


pdb_ligand_dictionary: Optional[Dict[str, str]] = None


def convert_to_molecule(residue: Residue,
                        template_mol: Optional[Mol] = None,
                        infer_bond_order: bool = True) -> Mol:
    with StringIO() as f:
        residue.save(f)
        residue_pdb_block = f.getvalue()

    mol = None

    # Try to apply bond orders from the PDB ligand entry

    if template_mol:
        pdb_mol = Chem.MolFromPDBBlock(residue_pdb_block)
        try:
            pdb_mol = AssignBondOrdersFromTemplate(template_mol, pdb_mol)
            if pdb_mol.HasSubstructMatch(template_mol):
                mol = pdb_mol
        except ValueError:
            pass

    # Otherwise let OpenBabel determine the bond orders

    if not mol and infer_bond_order:
        residue_mol_block = pybel.readstring(
            'pdb', residue_pdb_block).write('mol')
        mol = Chem.MolFromMolBlock(residue_mol_block)

    # If other attempts fail, return the molecule without bond orders

    if not mol:
        mol = Chem.MolFromPDBBlock(residue_pdb_block)

    return mol


def _get_pdb_ligand_dictionary() -> Dict[str, str]:
    global pdb_ligand_dictionary

    if not pdb_ligand_dictionary:
        url = 'http://ligand-expo.rcsb.org/dictionaries/' + \
            'Components-smiles-stereo-oe.smi'
        file_ligand_dictionary = get_file('Components-smiles-stereo-oe.smi',
                                          url)

        pdb_ligand_dictionary = dict()

        with open(file_ligand_dictionary) as file:
            csv_reader = csv.reader(file, delimiter='\t')
            for row in csv_reader:
                if len(row) > 1:
                    pdb_ligand_dictionary[row[1]] = row[0]

    return pdb_ligand_dictionary


def get_pdb_ligand(residue_name: str) -> Optional[Mol]:
    smiles = _get_pdb_ligand_dictionary().get(residue_name)
    return Chem.MolFromSmiles(smiles) if smiles else None


def read_oddt_molecule(filename: Union[str, Path]) -> oddt.toolkit.Molecule:
    if isinstance(filename, str):
        filename = Path(filename)

    mol = next(oddt.toolkit.readfile('mol', str(filename)))
    return mol



from rdkit.Chem import Mol, MolFromSmiles, MolToSmiles, SanitizeMol
from rdkit.Chem.AllChem import ReactionFromSmarts
from rdkit.Chem.MolStandardize import rdMolStandardize


class MolCleaner(object):
    def __init__(self):
        self._tautomerTetrazole = ReactionFromSmarts(
            '[*:1]-c1n[nH]nn1>>[*:1]-c1[nH]nnn1')

    def clean_mol(self, mol: Mol) -> Mol:
        """Cleans the specified molecule into standardised format.

        The steps are,

        - Removal of salts
        - Normalise structures
        - Normalise tautomers
        - Remove all charges (where possible)

        Args:
            mol: The molecule to clean.

        Returns:
            The cleaned molecule.
        """
        # Use RDKit standardizer to return the parent fragment (non-salt)
        # This will also apply more normalisation and clean up any charges

        mol = rdMolStandardize.ChargeParent(mol)

        # Custom tautomers

        mol = self._apply_reaction(mol, self._tautomerTetrazole)

        return mol

    def clean_smiles(self, smiles: str) -> str:
        """Cleans the specified SMILES into standardised format.

        The steps are,

        - Removal of salts
        - Normalise structures
        - Normalise tautomers
        - Remove all charges (where possible)

        Args:
            smiles: The SMILES to clean.

        Returns:
            The cleaned SMILES.
        """
        mol_input = MolFromSmiles(smiles)
        if not mol_input:
            return None
        mol_clean = self.clean_mol(mol_input)
        return MolToSmiles(mol_clean)

    def _apply_reaction(self, mol, reaction):
        products = reaction.RunReactants((mol,))
        if products:
            mol = products[0][0]
            SanitizeMol(mol)
            return mol
        else:
            return mol



import tensorflow as tf
from chemmltoolkit.tensorflow.utils import register_keras_custom_object


@register_keras_custom_object
class ShiftTensor(tf.keras.layers.Layer):
    """Shifts the input tensor along the time axis.
    This layer moves the contents of a tensor by the specified number of
    elements along the time axis (axis=1). Any elements that are moved
    outside the bounds of the tensor are dropped, with new elements padded
    with the specified value.
    Arguments:
        distance: A scalar defining the distance to shift the tensor.
        padding_value: A scalar defining the value to fill empty elements
            with. (default: 0)
    Input shape:
        nD tensor with shape: `(batch_size, time_axis, value_axis)`.
    Output shape:
        nD tensor with shape: `(batch_size, time_axis, value_axis)`.
    """

    def __init__(self,
                 distance,
                 padding_value=0,
                 **kwargs):
        super(ShiftTensor, self).__init__(**kwargs)
        self.distance = distance
        self.padding_value = padding_value

    def call(self, inputs):
        inputs = tf.convert_to_tensor(inputs)

        padding_amount = [self.distance, 0] if self.distance > 0 \
            else [0, -self.distance]
        padding_param = [padding_amount if i == 1 else [0, 0]
                         for i in range(len(inputs.shape))]
        padded = tf.pad(inputs,
                        padding_param,
                        constant_values=self.padding_value)

        if self.distance > 0:
            outputs = padded[:, :-self.distance]
        else:
            outputs = padded[:, -self.distance:]

        return outputs

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super(ShiftTensor, self).get_config()
        config.update({'distance': self.distance,
                       'padding_value': self.padding_value})
        return config



from chemmltoolkit.features.decorators import tokenizable_feature
from rdkit.Chem import Bond
from rdkit.Chem import BondStereo
from rdkit.Chem import BondType


@tokenizable_feature([BondType.SINGLE,
                      BondType.DOUBLE,
                      BondType.TRIPLE,
                      BondType.AROMATIC])
def bond_type(bond: Bond) -> BondType:
    """Type of bond (BondType).
    """
    return bond.GetBondType()


def index(bond: Bond) -> int:
    """Index within the parent molecule (int).
    """
    return bond.GetIdx()


def is_aromatic(bond: Bond) -> int:
    """If the bond is is_aromatic (0 or 1).
    """
    return int(bond.GetIsAromatic())


def is_conjugated(bond: Bond) -> int:
    """If the bond is conjugated (0 or 1).
    """
    return int(bond.GetIsConjugated())


def is_ring(bond: Bond) -> int:
    """If the bond is is in a ring (0 or 1).
    """
    return int(bond.IsInRing())


def is_ringsize(ringSize: int) -> int:
    """If the bond is is in a ring of the specified size (0 or 1).

    Args:
        ringSize: The size of the ring.
    """
    def _is_ringsize(bond: Bond):
        return int(bond.IsInRingSize(ringSize))
    _is_ringsize.__name__ = f'is_ringsize({ringSize})'
    return _is_ringsize


def order(bond: Bond) -> float:
    """Bond order (float).
    """
    return bond.GetBondTypeAsDouble()


@tokenizable_feature([BondStereo.STEREONONE,
                      BondStereo.STEREOANY,
                      BondStereo.STEREOZ,
                      BondStereo.STEREOE])
def stereochemistry(bond: Bond) -> BondStereo:
    """Stereochemistry (BondStereo).
    """
    return bond.GetStereo()



from typing import Callable, Optional, SupportsFloat
import numpy as np
from chemmltoolkit.data.quantities import Quantity
import math


def absolute(a: Quantity) -> Quantity:
    if a.val_min < 0 and a.val_max > 0:
        abs_min = -a.val_min
        if a.val_max > abs_min:
            return Quantity(0, a.val_max, False, a.eq_max, a.error)
        else:
            return Quantity(0, abs_min, False, a.eq_min, a.error)
    else:
        abs_min = float.__abs__(a.val_min)
        abs_max = float.__abs__(a.val_max)

        if abs_min < abs_max:
            return Quantity(abs_min, abs_max, a.eq_min, a.eq_max, a.error)
        else:
            return Quantity(abs_max, abs_min, a.eq_max, a.eq_min, a.error)


def add(a: Quantity, b: Quantity) -> Quantity:
    val_min = a.val_min + b.val_min
    val_max = a.val_max + b.val_max
    eq_min = a.eq_min and b.eq_min
    eq_max = a.eq_max and b.eq_max
    error = a.error + b.error

    return Quantity(val_min, val_max, eq_min, eq_max, error)


def greater(a: Quantity, b: Quantity) -> Optional[bool]:
    if a.val_min > b.val_max:
        return True
    elif (a.val_min == b.val_max) and (a.eq_min is False or b.eq_max is False):
        return True
    elif a.val_max <= b.val_min:
        return False
    return None


def greater_equal(a: Quantity, b: Quantity) -> Optional[bool]:
    if a.val_min >= b.val_max:
        return True
    elif a.val_max < b.val_min:
        return False
    elif (a.val_max == b.val_min) and (a.eq_max is False or b.eq_min is False):
        return False
    return None


def isclose(a: Quantity, b: Quantity,
            rel_tol: SupportsFloat = 1e-9,
            abs_tol: SupportsFloat = 0.0) -> bool:
    return math.isclose(a.val_min, b.val_min,
                        rel_tol=rel_tol, abs_tol=abs_tol) and \
        math.isclose(a.val_max, b.val_max,
                     rel_tol=rel_tol, abs_tol=abs_tol)


def less(a: Quantity, b: Quantity) -> Optional[bool]:
    if a.val_max < b.val_min:
        return True
    elif (a.val_max == b.val_min) and (a.eq_max is False or b.eq_min is False):
        return True
    elif a.val_min >= b.val_max:
        return False
    return None


def less_equal(a: Quantity, b: Quantity) -> Optional[bool]:
    if a.val_max <= b.val_min:
        return True
    elif a.val_min > b.val_max:
        return False
    elif (a.val_min == b.val_max) and (a.eq_min is False or b.eq_max is False):
        return False
    return None


def log10(x: Quantity) -> Quantity:
    return map_value(x, math.log10)


def map_value(x: Quantity, fn: Callable[[float], float]) -> Quantity:
    val_min = -np.inf if x.val_min == -np.inf else fn(x.val_min)
    val_max = np.inf if x.val_max == np.inf else fn(x.val_max)
    error = x.error if x.error in [0.0, np.inf] else fn(x.error)

    return Quantity(val_min, val_max, x.eq_min, x.eq_max, error)


def neg(a: Quantity) -> Quantity:
    return Quantity(-a.val_max, -a.val_min, a.eq_max, a.eq_min, a.error)


def pow(x: float, y: Quantity) -> Quantity:
    return map_value(y, lambda y: math.pow(x, y))


def sub(a: Quantity, b: Quantity) -> Quantity:
    return a + (-b)



import numpy as np


def add_master_node(adjacency_matrices):
    """Adds master nodes to adjacency matricies

    This will add a master node (an imaginary node that connects to all other
    nodes) to a set of adjacency matricies. The size of each adjacency matrix
    will hence increase by one in each dimension. Note that the master node
    is not connected to itself.

    Args:
        adjacency_matrices: The input list of adjacency matricies.

    Returns:
        The resulting list of adjacency matricies.
    """
    adj = np.pad(adjacency_matrices,
                 ((0, 0), (0, 1), (0, 1)),
                 mode='constant',
                 constant_values=1)
    adj[:, adj.shape[1]-1, adj.shape[2]-1] = 0
    return adj


def add_self_loops(adjacency_matrices):
    """Adds self loops to adjacency matricies

    This will add a self loops (connections between every node and itself) to
    a set of adjacency matricies.

    Args:
        adjacency_matrices: The input list of adjacency matricies.

    Returns:
        The resulting list of adjacency matricies.
    """
    adj = adjacency_matrices.copy()
    mask = np.eye(adj.shape[1], dtype=bool)
    adj[:, mask] = 1.0
    return adj


def normalise(adjacency_matrices):
    """Normalises adjacency matricies

    This will normalise (scale all rows to sum to one) a set of
    adjacency matricies.

    Args:
        adjacency_matrices: The input list of adjacency matricies.

    Returns:
        The resulting list of adjacency matricies.
    """
    def normalise_matrix(matrix):
        degree = np.sum(matrix, axis=0)
        degree_inv = np.power(degree, -1.0)
        inv_matrix = np.diag(degree_inv)
        return inv_matrix @ matrix

    adjs = [normalise_matrix(matrix) for matrix in adjacency_matrices]
    return np.array(adjs)


def normalise_spectral(adjacency_matrices):
    """Normalises adjacency matricies with a spectral method

    This will normalise a set of adjacency matricies using the spectral
    method of Kipf and Welling (https://arxiv.org/abs/1609.02907).

    Args:
        adjacency_matrices: The input list of adjacency matricies.

    Returns:
        The resulting list of adjacency matricies.
    """
    def normalise_matrix(matrix):
        degree = np.sum(matrix, axis=0)
        degree_inv = np.power(degree, -0.5)
        inv_matrix = np.diag(degree_inv)
        return inv_matrix @ matrix @ inv_matrix

    adjs = [normalise_matrix(matrix) for matrix in adjacency_matrices]
    return np.array(adjs)


def pad(adjacency_matrices, size):
    """Pads adjacency matricies to the desired size

    This will pad the adjacency matricies to the specified size, appending
    zeros as required. The output adjacency matricies will all be of size
    'size' x 'size'.

    Args:
        adjacency_matrices: The input list of adjacency matricies.
        size: The desired dimension of the output matricies.

    Returns:
        The resulting list of adjacency matricies.
    """
    padding = size - adjacency_matrices.shape[1]
    return np.pad(adjacency_matrices,
                  [(0, 0), (0, padding), (0, padding)],
                  mode='constant')



from typing import List
from chemmltoolkit.features.featuriser import Featuriser
from rdkit.Chem import MolFromSmiles
from rdkit.Chem import Mol


class MoleculeFeaturiser(Featuriser):
    """Generator for molecule-based features.

    Args:
        features: A list of features to generate.
    """
    def __init__(self, features: list):
        super(MoleculeFeaturiser, self).__init__(features)

    def process_molecule(self, mol: Mol):
        """Generates molecular features.

        Args:
            mol: The molecule to featurise.

        Returns:
            A list of features.
        """
        return self._process(mol)

    def get_feature_lengths(self) -> List[int]:
        """Calculates the length of each feature

        Returns:
            A list of the lengths of each feature.
        """
        mol = MolFromSmiles('CC')
        return self._get_feature_lengths(mol)



import tensorflow as tf
from chemmltoolkit.tensorflow.utils import register_keras_custom_object


@register_keras_custom_object
class OneHotEncoding(tf.keras.layers.Layer):
    """Applys one-hot encoding to the input.
    Arguments:
        depth: A scalar defining the depth of the one hot dimension.
        on_value: A scalar defining the value to fill in output when
            indices[j] = i. (default: 1)
        off_value: A scalar defining the value to fill in output when
            indices[j] != i. (default: 0)
        axis: The axis to fill (default: -1, a new inner-most axis).
        output_dtype: The data type of the output tensor.
    Input shape:
        nD tensor with shape: `(batch_size, ...)`.
    Output shape:
        nD tensor with shape: `(batch_size, ..., depth)`.
    """

    def __init__(self,
                 depth,
                 on_value=None,
                 off_value=None,
                 axis=None,
                 output_dtype=None,
                 **kwargs):
        super(OneHotEncoding, self).__init__(**kwargs)
        self.depth = depth
        self.on_value = on_value
        self.off_value = off_value
        self.axis = axis
        self.output_dtype = output_dtype

    def call(self, inputs):
        inputs = tf.convert_to_tensor(inputs)
        outputs = tf.one_hot(inputs, self.depth, self.on_value, self.off_value,
                             self.axis, self.output_dtype)
        return outputs

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()
        if self.axis:
            output_shape = input_shape[:self.axis] + \
                [self.depth] + \
                input_shape[self.axis:]
        else:
            output_shape = input_shape + [self.depth]
        return tf.TensorShape(output_shape)

    def get_config(self):
        config = super(OneHotEncoding, self).get_config()
        config.update({'depth': self.depth,
                       'on_value': self.on_value,
                       'off_value': self.off_value,
                       'axis': self.axis,
                       'output_dtype': self.output_dtype})
        return config



import tensorflow as tf


def determine_strategy(allow_cpu=False, allow_tpu=True):
    tpu = None
    if allow_tpu:
        try:
            tpu = tf.distribute.cluster_resolver.TPUClusterResolver()
        except ValueError:
            pass

    if tpu:
        tf.config.experimental_connect_to_cluster(tpu)
        tf.tpu.experimental.initialize_tpu_system(tpu)
        strategy = tf.distribute.experimental.TPUStrategy(tpu)
        strategy_type = 'tpu'
        print('Running on TPU ', tpu.cluster_spec().as_dict()['worker'])

    else:
        gpus = tf.config.experimental.list_logical_devices("GPU")
        if len(gpus) > 1:
            gpu_names = [gpu.name for gpu in gpus]
            strategy = tf.distribute.MirroredStrategy(gpu_names)
            strategy_type = 'gpu'
            print('Running on multiple GPUs ', gpu_names)
        elif len(gpus) == 1:
            strategy = tf.distribute.get_strategy()
            strategy_type = 'gpu'
            print('Running on single GPU ', gpus[0].name)
        else:
            if not allow_cpu:
                raise RuntimeError('TensorFlow hardware acceleration ' +
                                   'not found.')

            strategy = tf.distribute.get_strategy()
            strategy_type = 'cpu'
            print('Running on CPU')

    print("Number of accelerators: ", strategy.num_replicas_in_sync)
    return strategy, strategy_type



import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import rdMolDescriptors


class ConformerGenerator():
    """Generates conformers for molecules.

    Calling the 'generate_conformers' method will perform a number of steps.

    * Generate a number of candidate conformers, pruning any within a
      specified RMSD
    * Minimise the conformers with the specified forcefield
    * A further pruning step to remove conformers that have converged within
      the specified RMSD
    * Align the conformers to template molecules (if specified)

    Args:
        num_confs: The number of conformers to generate prior to minimisation.
            If this is not specified, a suitable number of conformers will
            be chosen per molecule based upon the flexibility.
        prune_rms_thresh: The RMS below which to prune similar conformers.
            Note that for speed the RMSD is calculated only for heavy atoms.
        embed_parameters: The RDKit EmbedParameters to use for generating
            conformers. By default this will be ETKDGv2. Note that for ease
            of use, the PruneRmsThresh property will always be set to the
            value of 'prune_rms_thresh'.
        force_field: The force field to use for minimisation. By default this
            will use UFF. Setting this value to 'None' will skip the
            minimisation step. Supported force fields are: 'uff', 'mmff94' or
            'mmff94s'.
        align_templates: A list of template molecules to align the generated
            conformers to. All matching template alignments are returned,
            however they are first pruned for similarity by
            'prune_rms_threshold' If this argument is not specified, no
            alignment is performed.
        align_rms_thresh: The RMS below which to accept alignments to
            template molecules.
    """

    def __init__(self,
                 num_confs=None,
                 prune_rms_thresh=0.35,
                 embed_parameters=None,
                 force_field='uff',
                 align_templates=None,
                 align_rms_thresh=1.0):
        self.num_confs = num_confs
        self.prune_rms_thresh = prune_rms_thresh
        self.force_field = force_field
        self.align_templates = align_templates
        self.align_rms_thresh = align_rms_thresh

        self.embed_parameters = embed_parameters if embed_parameters \
            else AllChem.ETKDGv2()
        self.embed_parameters.PruneRmsThresh = prune_rms_thresh

    def generate_conformers(self, mol):
        """Generates conformers for the specified molecule.

        Args:
            smiles: The input molecule.

        Returns:
            A new molecule with embedded conformers.
        """
        mol = Chem.AddHs(mol)
        mol = self._embed_conformers(mol)

        if self.force_field:
            energies = self._minimize_conformers(mol, self.force_field)
            mol = self._prune_conformers(mol, energies)

        if self.align_templates:
            mol = self._align_conformers(mol, self.align_templates)

        return mol

    def _embed_conformers(self, mol):
        if self.num_confs:
            num_confs = self.num_confs
        else:
            n_rot = rdMolDescriptors.CalcNumRotatableBonds(mol)
            if n_rot <= 7:
                num_confs = 50
            elif n_rot <= 12:
                num_confs = 200
            else:
                num_confs = 300

        AllChem.EmbedMultipleConfs(mol,
                                   numConfs=num_confs,
                                   params=self.embed_parameters)

        return mol

    def _minimize_conformers(self, mol, force_field):
        if force_field == 'uff':
            result = AllChem.UFFOptimizeMoleculeConfs(mol)
        elif force_field.startswith('mmff'):
            result = AllChem.MMFFOptimizeMoleculeConfs(mol,
                                                       mmffVariant=force_field)
        else:
            raise ValueError(f'Unexpected force field {force_field}')

        for conformer, (success, _) in zip(mol.GetConformers(), result):
            if success != 0:
                mol.RemoveConformer(conformer.GetId())

        return [energy for success, energy in result if success == 0]

    def _prune_conformers(self, mol, energies):
        # Only keep the lowest energy conformers within each RMSD
        # Uses the procedure of Chem. Inf. Model., 2012, 52, 1146
        conformers = mol.GetConformers()
        confIds_sorted = [conformers[int(i)].GetId()
                          for i in np.argsort(energies)]
        confIds_keep = []
        mol_noH = Chem.RemoveHs(mol)

        for confId in confIds_sorted:
            keep_conformer = True
            for refId in confIds_keep:
                rmsd = AllChem.GetBestRMS(mol_noH, mol_noH,
                                          prbId=confId, refId=refId)
                if rmsd < self.prune_rms_thresh:
                    keep_conformer = False
                    break

            if keep_conformer:
                confIds_keep.append(confId)

        # Return a new molecule with only the chosen conformers
        result = Chem.Mol(mol)
        result.RemoveAllConformers()

        for confId in confIds_keep:
            result.AddConformer(mol.GetConformer(confId), assignId=True)

        return result

    def _align_conformers(self, mol, templates):
        def _get_maps(mol, template):
            matches = mol.GetSubstructMatches(template)
            match_template = template.GetSubstructMatch(template)
            return [(template, list(zip(m, match_template))) for m in matches]

        template_maps = [_get_maps(mol, template) for template in templates]
        template_maps = [item for list in template_maps for item in list]
        template_maps = sorted(template_maps,
                               key=lambda x: len(x[1]),
                               reverse=True)

        result = Chem.Mol(mol)
        result.RemoveAllConformers()

        for conformer in mol.GetConformers():
            candidate_conformers = []
            for template, constraint_map in template_maps:
                score = AllChem.AlignMol(mol,
                                         template,
                                         prbCid=conformer.GetId(),
                                         atomMap=constraint_map)

                if score <= self.align_rms_thresh:
                    newConfId = result.AddConformer(conformer, assignId=True)
                    candidate_conformers.append(newConfId)

            accepted_conformers = candidate_conformers[:1]
            for candidate_confId in candidate_conformers[1:]:
                rmsds = [AllChem.GetConformerRMS(result,
                                                 candidate_confId,
                                                 accepted_confId,
                                                 prealigned=True)
                         for accepted_confId in accepted_conformers]
                if min(rmsds) > self.prune_rms_thresh:
                    accepted_conformers.append(candidate_confId)
                else:
                    result.RemoveConformer(candidate_confId)

        return result



from typing import Any, List
from chemmltoolkit.features.featuriser import Featuriser
from rdkit.Chem import MolFromSmiles
from rdkit.Chem import Mol
from rdkit.Chem import Atom


class AtomFeaturiser(Featuriser):
    """Generator for atom-based features.
    Args:
        features: A list of features to generate.
    """

    def __init__(self, features: list):
        super(AtomFeaturiser, self).__init__(features)

    def process_atom(self, atom: Atom):
        """Generates features for an individual atom.

        Args:
            atom: The atom to featurise.

        Returns:
            A list of features.
        """
        return self._process(atom)

    def process_molecule(self, mol: Mol) -> List[List[Any]]:
        """Generates features for all atoms in a molecule.

        Args:
            mol: The molecule to featurise.

        Returns:
            A nested list of features for all atoms.
        """
        atoms = mol.GetAtoms()
        return [self._process(atom) for atom in atoms]

    def get_feature_lengths(self) -> List[int]:
        """Calculates the length of each feature

        Returns:
            A list of the lengths of each feature.
        """
        molecule = MolFromSmiles('CC')
        atom = molecule.GetAtoms()[0]
        return self._get_feature_lengths(atom)



from typing import List, Optional, Tuple

from Bio.PDB.Atom import Atom
from . import Chain
from Bio import pairwise2
from Bio.PDB import Superimposer
from Bio.PDB.Residue import Residue as bpResidue


def align(mobile: Chain,
          reference: Chain
          ) -> Tuple[Optional[Chain], int, Optional[float]]:
    align_mobile, align_ref = _get_residue_alignments(mobile, reference)
    num_align_residues = len(align_mobile)
    if num_align_residues == 0:
        return None, 0, None

    atoms_mobile = [residue.child_dict.get('CA') for residue in align_mobile]
    atoms_ref = [residue.child_dict.get('CA') for residue in align_ref]

    atoms_paired = [(a, b) for (a, b) in zip(
        atoms_mobile, atoms_ref) if a is not None and b is not None]
    if len(atoms_paired) == 0:
        return None, 0, None
    atoms_mobile, atoms_ref = zip(*atoms_paired)

    superimposer = Superimposer()
    superimposer.set_atoms(atoms_ref, atoms_mobile)

    all_atoms_disordered: List[Atom] = []
    for residue in mobile.as_biopython().get_residues():
        # NB: Get all atoms including disordered residues
        for atom in residue.get_unpacked_list():
            all_atoms_disordered.append(atom)

    superimposer.apply(all_atoms_disordered)

    return mobile, num_align_residues, superimposer.rms


def _get_residue_alignments(chainA: Chain,
                            chainB: Chain
                            ) -> Tuple[List[bpResidue], List[bpResidue]]:
    def _seq_to_residues(chain: Chain, sequence: str):
        residues = list(chain.as_biopython().get_residues())
        offset = 0

        def _get_next_residue():
            nonlocal offset
            residue = residues[offset]
            offset += 1
            return residue

        return [None if code in ['-', '?'] else _get_next_residue()
                for code in sequence]

    seqA = chainA.get_sequence()
    seqB = chainB.get_sequence()

    alignment = pairwise2.align.globalxx(seqA, seqB)[0]

    residuesA = _seq_to_residues(chainA, alignment.seqA)
    residuesB = _seq_to_residues(chainB, alignment.seqB)

    alignA: List[bpResidue] = []
    alignB: List[bpResidue] = []

    for resA, resB in zip(residuesA, residuesB):
        if resA is not None and resB is not None:
            alignA.append(resA)
            alignB.append(resB)

    return alignA, alignB



import tensorflow as tf
from tensorflow import DType

NODE_FEATURES = 'node_features'
EDGE_FEATURES = 'edge_features'


def TensorGraph(node_features, edge_features):
    """Returns a tensor graph instance.

    This function simply returns a dictionary with the relevant structure
    to represent a graph of Tensorflow Tensors.

    Arguments:
        node_features: A tensor of node features.
        edge_features: A tensor of edge features.
    """
    return {
        NODE_FEATURES: node_features,
        EDGE_FEATURES: edge_features
    }


class TensorGraphShape:
    """Represents the shape of a TensorGraph.

    The shape of a TensorGraph encompases all the information that describes
    the graph's underlying tensors. This includes what tensors are specified,
    their dimensions, as well as how they are represented.

    Arguments:
        num_nodes: The number of nodes in the graph, or None if unknown.
        node_dims: The number of features assigned to each node.
        edge_dims: The number of features assigned to each edge.
        node_dtype: The TensorFlow datatype for the node features.
        edge_dtype: The TensorFlow datatype for the edge features.
    """

    def __init__(self,
                 num_nodes: int = None,
                 node_dims: int = 1, edge_dims: int = 1,
                 node_dtype: DType = tf.float32,
                 edge_dtype: DType = tf.float32):
        self.num_nodes = num_nodes
        self.node_dims = node_dims
        self.edge_dims = edge_dims
        self.node_dtype = node_dtype
        self.edge_dtype = edge_dtype

    def get_padding_shapes(self):
        return {
            NODE_FEATURES: (self.num_nodes, self.node_dims),
            EDGE_FEATURES: (self.edge_dims, self.num_nodes, self.num_nodes)
        }

    def get_padding_values(self):
        return {
            'node_features': 0.0,
            'edge_features': 0.0
        }


def map_node_features(graph, map_fn):
    features_in = graph[NODE_FEATURES]
    features_out = map_fn(features_in)
    return {**graph, NODE_FEATURES: features_out}


def map_edge_features(graph, map_fn):
    features_in = graph[EDGE_FEATURES]
    features_out = map_fn(features_in)
    return {**graph, EDGE_FEATURES: features_out}



from .protein import Protein, Chain, Residue
from .protein import ResidueType
from .alignment import align
from .interactions import get_covalent_residue, contact_distance
from .interactions import get_interactions
from .molecule import convert_to_molecule, get_pdb_ligand, read_oddt_molecule

__all__ = ['Protein', 'Chain', 'Residue',
           'ResidueType',
           'align',
           'get_covalent_residue', 'contact_distance',
           'get_interactions',
           'convert_to_molecule', 'get_pdb_ligand', 'read_oddt_molecule']



def normalizable_feature(mean, std):
    """Decorator for features to specify default normalization.

    Args:
        mean: The mean value for the feature.
        std: The standard deviation for the feature.
    """
    def _normalizable_feature(func):
        func.normal_mean = mean
        func.normal_std = std
        return func
    return _normalizable_feature


def tokenizable_feature(tokens):
    """Decorator for features to specify default tokens.

    Args:
        tokens: The default set of tokens for the feature.
    """
    def _tokenizable_feature(func):
        func.tokens = tokens
        return func
    return _tokenizable_feature



import tensorflow as tf


class VariableScheduler(tf.keras.callbacks.Callback):
    """Schedules an arbitary variable during training.
    Arguments:
        variable: The variable to modify the value of.
        schedule:  A function that takes an epoch index (integer, indexed
            from 0) and current variable value as input and returns a new
            value to assign to the variable as output.
        verbose: int. 0: quiet, 1: update messages.
    """

    def __init__(self, variable, schedule, verbose=0):
        super(VariableScheduler, self).__init__()
        self.variable = variable
        self.schedule = schedule
        self.verbose = verbose

    def on_epoch_begin(self, epoch, logs=None):
        value = self.variable.read_value()
        value = self.schedule(epoch, value)
        self.variable.assign(value)
        if self.verbose > 0:
            print(f'\nEpoch {epoch + 1}: VariableScheduler assigning '
                  f'variable {self.variable.name} to {value}.')



from chemmltoolkit.tensorflow.losses.mean_absolute_error_nan \
    import mean_absolute_error_nan, mae_nan, MAE_nan, MeanAbsoluteErrorNaN
from chemmltoolkit.tensorflow.losses.mean_squared_error_nan \
    import mean_squared_error_nan, mse_nan, MSE_nan, MeanSquaredErrorNaN

__all__ = ['mean_absolute_error_nan', 'mae_nan', 'MAE_nan',
           'MeanAbsoluteErrorNaN',
           'mean_squared_error_nan', 'mse_nan', 'MSE_nan',
           'MeanSquaredErrorNaN']



import abc
import tensorflow as tf
from tensorflow.keras.metrics import Metric


class MeanMetricNaN(Metric):
    """ An abstract base class for metrics calculated by a mean of some error
    values. It handles the case where some 'y_true' values are NaN.
    Additionally, a 'per_sample' parameter will average over samples as well
    as overall.

    Derived classes must implement 'calculate_error' to return a calculated
    error from the true and predicted values. Optionally they can also
    implement 'transform_mean' where some transformation is applied to the
    mean values to determine the metric (this transformation is applied
    globally, or per sample based on the value of 'per_sample').
    """

    def __init__(self, per_sample=False, name='metric_nan', dtype=tf.float32):
        super(MeanMetricNaN, self).__init__(name=name, dtype=dtype)
        self.per_sample = per_sample
        self.total = self.add_weight('total', initializer='zeros')
        self.count = self.add_weight('count', initializer='zeros')

    @abc.abstractmethod
    def calculate_error(self, y_true, y_pred):
        raise NotImplementedError('Must be implemented in subclasses.')

    def transform_mean(self, mean):
        return mean

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true = tf.convert_to_tensor(y_true, self.dtype)
        y_pred = tf.convert_to_tensor(y_pred, self.dtype)

        error = self.calculate_error(y_true, y_pred)
        mask = ~tf.math.is_nan(error)
        error = tf.ragged.boolean_mask(error, mask)

        if self.per_sample:
            error = self.transform_mean(tf.reduce_mean(error, axis=-1))
            count = tf.cast(~tf.math.is_nan(error), tf.float32)
        else:
            error = tf.reduce_sum(error, axis=-1)
            count = tf.reduce_sum(tf.cast(mask, tf.float32), axis=-1)

        if sample_weight is not None:
            sample_weight = tf.cast(sample_weight, self.dtype)
            error = error * sample_weight
            count = count * sample_weight

        error = tf.boolean_mask(error, ~tf.math.is_nan(error))

        self.total.assign_add(tf.reduce_sum(error))
        self.count.assign_add(tf.reduce_sum(count))

    def result(self):
        mean = self.total / self.count

        if self.per_sample:
            return mean
        else:
            return self.transform_mean(mean)

    def reset_states(self):
        self.total.assign(0.0)
        self.count.assign(0.0)


class MeanAbsoluteErrorNaN(MeanMetricNaN):
    """Computes the mean absolute error, masking any NaN values in `y_true`.
    Arguments:
        per_sample: If false (the default), the error is calculated over all
            individual values. If true, the error is calculated for each
            sample individually, and the overall mean of these errors is
            returned.
    """
    def __init__(self,
                 per_sample=False,
                 name='mean_absolute_error_nan',
                 dtype=tf.float32):
        super(MeanAbsoluteErrorNaN, self).__init__(
            per_sample=per_sample, name=name, dtype=dtype)

    def calculate_error(self, y_true, y_pred):
        return tf.abs(y_pred - y_true)


class MeanSquaredErrorNaN(MeanMetricNaN):
    """Computes the mean squared error, masking any NaN values in `y_true`.
    Arguments:
        per_sample: If false (the default), the error is calculated over all
            individual values. If true, the error is calculated for each
            sample individually, and the overall mean of these errors is
            returned.
    """
    def __init__(self,
                 per_sample=False,
                 name='mean_squared_error_nan',
                 dtype=tf.float32):
        super(MeanSquaredErrorNaN, self).__init__(
            per_sample=per_sample, name=name, dtype=dtype)

    def calculate_error(self, y_true, y_pred):
        return (y_pred - y_true) ** 2


class RootMeanSquaredErrorNaN(MeanMetricNaN):
    """Computes the root mean squared error, masking any NaN values in
    `y_true`.
    Arguments:
        per_sample: If false (the default), the error is calculated over all
            individual values. If true, the error is calculated for each
            sample individually, and the overall mean of these errors is
            returned.
    """
    def __init__(self,
                 per_sample=False,
                 name='root_mean_squared_error_nan',
                 dtype=tf.float32):
        super(RootMeanSquaredErrorNaN, self).__init__(
            per_sample=per_sample, name=name, dtype=dtype)

    def calculate_error(self, y_true, y_pred):
        return (y_pred - y_true) ** 2

    def transform_mean(self, mean):
        return tf.sqrt(mean)



import tensorflow as tf
from chemmltoolkit.tensorflow.graph import TensorGraph
from chemmltoolkit.tensorflow.graph import TensorGraphShape


def define_tfrecord_features(shape: TensorGraphShape, name: str = 'graph'):
    return {
        f'{name}:num_nodes': tf.io.FixedLenFeature([], tf.int64),
        f'{name}:node_features': tf.io.VarLenFeature(shape.node_dtype),
        f'{name}:edge_features': tf.io.VarLenFeature(shape.edge_dtype)
    }


def tfrecord_to_tensorgraph(example: dict, name: str = 'graph'):
    num_nodes = example[name + ':num_nodes']

    node_features = tf.sparse.to_dense(example[name + ':node_features'])
    node_features = tf.reshape(node_features, (num_nodes, -1))

    edge_features = tf.sparse.to_dense(example[name + ':edge_features'])
    edge_features = tf.reshape(edge_features, (-1, num_nodes, num_nodes))

    return TensorGraph(node_features, edge_features)



import pathlib
from setuptools import setup, find_packages
from chemmltoolkit.__version__ import __version__

# The directory containing this file
here = pathlib.Path(__file__).parent

# Read files for inclusion in the setup
readme = (here / "README.md").read_text()
license = (here / "LICENSE").read_text()

# This call to setup() does all the work
setup(
    name="chemmltoolkit",
    version=__version__,
    description="Useful functionality for machine learning in chemistry",
    long_description=readme,
    long_description_content_type="text/markdown",
    license=license,
    url="https://github.com/Andy-Wilkinson/ChemMLToolkit",
    author="Andrew Wilkinson",
    packages=find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
)



from typing import Any, List
from chemmltoolkit.utils.list_utils import flatten
from chemmltoolkit.utils.list_utils import _is_iterable
from chemmltoolkit.features.utils import get_feature_keys


class Featuriser:
    """A generic feature generator.

    This will generally be the base class for specific featurisers.

    Args:
        features: A list of features to generate.
    """
    def __init__(self, features: list):
        self.features = features

    def _process(self, data) -> List[Any]:
        """Generates features for an individual data point.

        Args:
            data: The data to featurise.

        Returns:
            A list of features.
        """

        features = [feature(data) for feature in self.features]
        return flatten(features)

    def _get_feature_lengths(self, example) -> List[int]:
        """Calculates the length of each feature

        This uses an example data item, generate each feature in turn
        and return a list of each length.

        Args:
            data: Example data to featurise.

        Returns:
            A list of the lengths of each feature.
        """
        def _get_feature_length(feature):
            val = feature(example)
            if _is_iterable(val):
                return len(val)
            else:
                return 1

        return [_get_feature_length(feature) for feature in self.features]

    def get_feature_names(self) -> List[str]:
        """Gets strings representing each type of feature

        This will return a list of strings, one for each feature function
        specified when creating the featuriser. This can be used for example
        to log the features chosen, or to save them for recreation later.

        Returns:
            A list of feature descriptions.
        """
        return flatten([feature.__name__ for feature in self.features])

    def get_feature_keys(self) -> List[str]:
        """Gets the name of each element in the generated feature

        This will return a list of strings, one for each output value when
        generating features. You can use this to display a human-readable
        description of each element of the feature (e.g. when understanding
        feature importance). This will typically be the name of the feature,
        with some form of index appended if the feature outputs multiple
        values.

        Returns:
            A list of feature names.
        """
        lengths = self.get_feature_lengths()

        return flatten([get_feature_keys(feature, length)
                        for feature, length
                        in zip(self.features, lengths)])



import importlib
import os
import stat
import subprocess
import sys


from chemmltoolkit.utils.data_utils import get_file


def install_conda_package(package_spec,
                          channel=None,
                          conda_location='/usr/local/'):
    """Installs conda packages.

    Note that this will install packages into the current conda environment.
    In many cases this will be the 'base' environment.

    Args:
        package_spec: The package to install.
        channel: The channel to use (or the default channel if None).
        conda_location: The location in which conda is installed.
    """
    conda_args = ['install', '-q', '-y', package_spec]
    if channel:
        conda_args = conda_args + ['-c', channel]

    conda_bin = os.path.join(conda_location, 'condabin/conda')
    subprocess.run([conda_bin] + conda_args)


def install_miniconda(installer_url='https://repo.anaconda.com/miniconda/' +
                                    'Miniconda3-latest-Linux-x86_64.sh',
                      conda_location='/usr/local/'):
    """ Install Miniconda if it does not already exist.

    Args:
        installer_url: The URL to the Miniconda installer.
        conda_location: The location to install Miniconda.
    """
    if os.path.exists(os.path.join(conda_location, 'condabin')):
        import conda
        print(f'Conda version {conda.__version__} already installed')
    else:
        installer_file = get_file('Miniconda3-latest-Linux-x86_64.sh',
                                  installer_url)
        os.chmod(installer_file,
                 os.stat(installer_file).st_mode | stat.S_IEXEC)
        subprocess.run([installer_file, f'-b -f -p {conda_location}'])
        sys.path.append(os.path.join(conda_location,
                                     'lib/python3.7/site-packages/'))
        import conda
        print(f'Conda version {conda.__version__} installed successfully')


def install_rdkit():
    """Installs RDKit.

    This function will install RDKit using conda (installing Miniconda first
    if required)
    """
    if importlib.util.find_spec('rdkit'):
        import rdkit
        print(f'RDKit version {rdkit.__version__} already installed')
    else:
        install_miniconda()
        install_conda_package('rdkit', channel='rdkit')

        # Check installation success

        import rdkit
        print(f'RDKit version {rdkit.__version__} installed successfully')



from chemmltoolkit.utils import list_utils
from chemmltoolkit.features.utils import get_token_name


def normalize(feature, mean=None, std=None):
    """Wraps a feature to normalize the value

    This returns a new feature that first calls the specified feature, then
    returns the result normalized by the transformation '(value - mean) / std'.

    If the 'mean' or 'std' parameters are not specified, the default values
    for the feature are infered from a 'normalizable_feature' decorator if
    present.

    Args:
        feature: The input feature to normalize.
        mean: The mean value for the feature.
        std: The standard deviation for the feature.

    Returns:
        The normalized feature.
    """
    if mean is None:
        mean = feature.normal_mean
    if std is None:
        std = feature.normal_std

    def _normalize(input):
        return (feature(input) - mean) / std

    _normalize.__name__ = \
        f'normalize({feature.__name__}, mean={mean}, std={std})'
    return _normalize


def one_hot(feature, tokens=None):
    """Wraps a feature to one-hot encode the value

    This returns a new feature that first calls the specified feature, then
    returns the result as a one-hot encoding of the specified token list.

    If the 'tokens' parameter is not specified, the default value for the
    feature is infered from a 'tokenizable_feature' decorator if present.

    Args:
        feature: The input feature to one-hot encode.
        tokens: A list of tokens to use for encoding.

    Returns:
        The one-hot encoded feature.
    """
    if tokens is None:
        tokens = feature.tokens

    feature_name = feature.__name__
    token_names = [get_token_name(token) for token in tokens]

    def _one_hot(input):
        return list_utils.one_hot(feature(input), tokens)

    def _get_feature_keys():
        return [f'one_hot({feature_name})[{token_name}]'
                for token_name in token_names]

    token_name_list = ','.join(token_names)
    _one_hot.__name__ = f'one_hot({feature_name}, tokens=[{token_name_list}])'
    _one_hot.get_feature_keys = _get_feature_keys
    return _one_hot
